unit BTFileStream;

{ * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
  * BTFileStream                                                            *
  *- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*
  * Copyright (c) 2002 - 2010 by Martin Offenwanger (www.dsplayer.de)  	    *
  *- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*
  * Mozilla Public License Version 1.1:                                     *
  *- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*
  * The contents of this file are used with permission, subject to the      *
  * Mozilla Public License Version 1.1 (the "License"); you may             *
  * not use this file except in compliance with the License. You may        *
  * obtain a copy of the License at                                         *
  * http://www.mozilla.org/MPL/MPL-1.1.html                                 *
  *                                                                         *
  * Software distributed under the License is distributed on an             *
  * "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or          *
  * implied. See the License for the specific language governing            *
  * rights and limitations under the License.                               *
  * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * }
{
  @author(Martin Offenwanger: coder@dsplayer.de)
  @created(May 16, 2007)
  @lastmod(Dec 25, 2010)
  @supported compilers: Delphi 7 - DelphiXE
}
{$WARN UNSAFE_CODE OFF}
{$WARN UNSAFE_TYPE OFF}
{$WARN UNSAFE_CAST OFF}

interface

/// /////////////////////////////////////////////////////////////////////////////
// ------------------------------------------------------------------------------
// Windows API based synchron File Stream Class
// -----------------------------------------------------------------------------
// Features:
// * 32- and 64bit filesizes
// * read-write network files
// * Ansi and Unicode File names
// * class error handling
// * system error handling
// ------------------------------------------------------------------------------
/// /////////////////////////////////////////////////////////////////////////////

/// /////////////////////////////////////////////////////////////////////////////
// ------------------------------------------------------------------------------
// Motivation of this unit:
// No more need to use Delphi's VCL TFilestream class (buggy sharing mode)
// and provide a more extended error handling close to the WinAPI
// ------------------------------------------------------------------------------
/// /////////////////////////////////////////////////////////////////////////////

/// /////////////////////////////////////////////////////////////////////////////
// ******************************************************************************
// Supported OS: Win2000, WinXP, WinVista, Windows7
// ******************************************************************************
uses BTStd, // standard include of all BT project codes
  Windows; // ....sorry Linux freaks, this is pure windows stuff ;P ....
// ******************************************************************************
/// /////////////////////////////////////////////////////////////////////////////

/// /////////////////////////////////////////////////////////////////////////////
// Class error strings (returned by GetErrorString)                        BEGIN
// ------------------------------------------------------------------------------
// -> this errors strings will occour, if you use TBTFileStream in a wrong way
// ------------------------------------------------------------------------------
// ATTENTION: you will not find a declaration for system error strings or
// translated sytem error strings in this unit. They will be automaticly
// generated in the current OS language and also returned by using
// GetErrorString. Call GetErrorString always immediately after the
// failed method call to get a more specified system error information string.
// ------------------------------------------------------------------------------
const
  BT_FS_ERRSTR_INVALID_POSITION = 'You are trying to access a invalid ' +
    'stream position which is smaller 0 or bigger as the current ' +
    'opened filestream.';
  BT_FS_ERRSTR_FILE_IS_OPEN = 'There is still a file opened. Call CloseFile ' +
    'before opening a new file.';
  BT_FS_ERRSTR_FILE_IS_NOT_OPEN = 'You need to perform a successful ' +
    'OpenFile call before.';
  BT_FS_ERRSTR_INVALID_FILE_OPEN_MODE = 'You are trying to set an invalid ' +
    'file open mode. Do only use: btFmOpenRead, btFmOpenWrite or ' +
    'btFmOpenReadWrite for it.';
  BT_FS_ERRSTR_INVALID_FILE_OPEN_PARAM = 'You are trying to set an invalid ' +
    'file open mode parameter. Do only use: btFmpCreateAlways, ' +
    'btFmpCreateNeweIfNotExist, btFmpOpenAlways, btFmpOpenExisting or ' +
    'btFmpTrunkateIfExist for it.';
  BT_FS_ERRSTR_INVALID_FILE_SHARE_MODE = 'You are trying to set an invalid ' +
    'file share mode. Do only use: btfmShareWrite, btfmShareRead, ' +
    'btfmShareReadWrite or btfmShareNone for it.';
  BT_FS_ERRSTR_FILE_WRITE_NOT_SUPPORTED = 'Write to file is not supported, ' +
    'because file open mode was set to read only. For write access set the ' +
    'file mode to btFmOpenWrite or btFmOpenReadWrite. ';
  BT_FS_ERRSTR_FILE_READ_NOT_SUPPORTED = 'Read from file is not supported, ' +
    'because file open mode was set to write only. For read access set the ' +
    'file mode to btFmOpenRead or btFmOpenReadWrite. ';
  BT_FS_ERRSTR_UNKNOWN = 'An unknown error occured';
  // ------------------------------------------------------------------------------
  // Class error strings (returned by GetErrorString)                          END
  /// /////////////////////////////////////////////////////////////////////////////

  /// /////////////////////////////////////////////////////////////////////////////
  // -> FILE OPEN MODE <-                                                   Begin
  // ------------------------------------------------------------------------------
  { Type definition: } type
  TBTFileOpenMode = type DWORD;

const
  // ------------------------------------------------------------------------------
  btFmOpenRead = $00000000; // opens the file for read access
  btFmOpenWrite = $00000001; // opens the file for write access
  btFmOpenReadWrite = $00000002; // opens the file for read-write access
  // ------------------------------------------------------------------------------
  // -> FILE OPEN MODE <-                                                    End
  /// /////////////////////////////////////////////////////////////////////////////

  /// /////////////////////////////////////////////////////////////////////////////
  // -> FILE OPEN MODE PARAMETER <-                                          Begin
  // ------------------------------------------------------------------------------
  { Type definition: } type
  TBTFileModeParam = type DWORD;

const
  // ------------------------------------------------------------------------------
  btFmpCreateAlways = $00000003; // create always new file
  btFmpCreateNeweIfNotExist = $00000004; (* create new file if not exist.
    If file exist OpenFile fails *)
  btFmpOpenAlways = $00000005; (* open a existing file. If a file
    do not exist, then create one *)
  btFmpOpenExisting = $00000006; // open a existing file
  btFmpTrunkateIfExist = $00000007; (* open only a existing file and
    set them it to size 0 *)
  // ------------------------------------------------------------------------------
  // -> FILE OPEN MODE PARAMETER <-                                            End
  /// /////////////////////////////////////////////////////////////////////////////

  /// /////////////////////////////////////////////////////////////////////////////
  // -> FILE SHARE MODE <-                                                   Begin
  // ------------------------------------------------------------------------------
  { Type definition: } type
  TBTFileShareMode = type DWORD;

const
  // ------------------------------------------------------------------------------
  btfmShareWrite = $00000008; // Share file for write access
  btfmShareRead = $00000009; // Share file for read access
  btfmShareReadWrite = $0000000A; // Share file for read-write access
  btfmShareNone = $0000000B; // Share file for nothing
  // ------------------------------------------------------------------------------
  // -> FILE SHARE MODE <-                                                     End
  /// /////////////////////////////////////////////////////////////////////////////

type
  TBTFileStream = class(TObject)
  private
    m_FileIsOpen: Boolean;
    m_FileHandle: THandle;
    m_LastWinAPIError: DWORD;
    m_LastErrorWasSystemError: Boolean;
    m_OpenMode: TBTFileOpenMode;
    function SystemErrorToBTError(f_SystemError: DWORD): TBTCommonError;
    function OpenFileEx(const f_FilenameAnsi: PAnsiChar;
      const f_FilenameUnicode: PWideChar;
      const f_OpMode: TBTFileOpenMode; const f_OpParam: TBTFileModeParam;
      const f_ShMode: TBTFileShareMode = btfmShareNone): TBTCommonError;
  public
    constructor Create;
    destructor Destroy; override;

    /// /////////////////////////////////////////////////////////////////////////
    // OpenFile:        Opens a File for reading and/or writing
    // -------------------------------------------------------------------------
    // Parameters:      f_OpMode:  btFmOpenRead, btFmOpenWrite,
    // btFmOpenReadWrite
    // f_OpParam: btFmpCreateAlways, btFmpCreateNeweIfNotExist,
    // btFmpOpenAlways, btFmpOpenExisting,
    // btFmpTrunkateIfExist
    // f_ShMode:  btfmShareWrite, btfmShareRead,
    // btfmShareReadWrite, btfmShareNone
    // -------------------------------------------------------------------------
    // Errors:          BT_FILE_ERROR_FILE_IS_OPEN,
    // BT_FILE_ERROR_INVALID_FILE_OPEN_MODE,
    // BT_FILE_ERROR_INVALID_FILE_OPEN_PARAM,
    // BT_FILE_ERROR_INVALID_FILE_SHARE_MODE,
    // BT_FILE_ERROR_FILE_NOT_FOUND,
    // BT_FILE_ERROR_PATH_NOT_FOUND,
    // BT_FILE_ERROR_TOO_MANY_OPEN_FILES,
    // BT_FILE_ERROR_WRITE_PROTECT,
    // BT_FILE_ERROR_SHARING_VIOLATION,
    // BT_FILE_ERROR_SHARING_BUFFER_EXCEEDED,
    // BT_FILE_ERROR_LOCK_VIOLATION,
    // BT_FILE_ERROR_FILE_ALREADY_EXISTS,
    // BT_FILE_ERROR_OPEN_FAILED,
    // BT_FILE_ERROR_NAME_TOO_LONG,
    // BT_FILE_ERROR_HANDLE_DISK_FULL,
    // BT_FILE_ERROR_DISK_FULL,
    // BT_FILE_ERROR_NO_MORE_HANDLES,
    // BT_FILE_ERROR_INVALID_NAME,
    // BT_FILE_ERROR_ALREADY_EXISTS_ON_CREATE,
    // BT_FILE_ERROR_FILENAME_EXCED_RANGE,
    // BT_FILE_ERROR_META_EXPANSION,
    // BT_FILE_ERROR_FILE_CORRUPT,
    // BT_FILE_ERROR_DISK_CORRUPT,
    // BT_FILE_ERROR_ACCESS_DENIED,
    // BT_NETWORK_ERROR_NETWORK_NOT_SUPPORTED,
    // BT_NETWORK_ERROR_BAD_NETPATH,
    // BT_NETWORK_ERROR_NETWORK_BUSY,
    // BT_ERROR_SYSTEM
    // -------------------------------------------------------------------------
    function OpenFile(const f_Filename: string;
      const f_OpMode: TBTFileOpenMode; const f_OpParam: TBTFileModeParam;
      const f_ShMode: TBTFileShareMode = btfmShareNone): TBTCommonError;

    function OpenFileA(const f_Filename: AnsiString;
      const f_OpMode: TBTFileOpenMode; const f_OpParam: TBTFileModeParam;
      const f_ShMode: TBTFileShareMode = btfmShareNone): TBTCommonError;

    function OpenFileW(const f_Filename: WideString;
      const f_OpMode: TBTFileOpenMode; const f_OpParam: TBTFileModeParam;
      const f_ShMode: TBTFileShareMode = btfmShareNone): TBTCommonError;

    /// /////////////////////////////////////////////////////////////////////////

    /// /////////////////////////////////////////////////////////////////////////
    // SetPosition:     Set the file pointer position
    // -------------------------------------------------------------------------
    // Parameters:      f_pos: new file pointer position
    // -------------------------------------------------------------------------
    // Errors:          BT_FILE_ERROR_FILE_IS_NOT_OPEN,
    // BT_FILE_ERROR_INVALID_POSITION,
    // BT_ERROR_SYSTEM *)
    function SetPosition(const f_pos: int64): TBTCommonError;
    /// /////////////////////////////////////////////////////////////////////////

    /// /////////////////////////////////////////////////////////////////////////
    // GetPosition:     Get the current file pointer position
    // -------------------------------------------------------------------------
    // Parameters:      f_pos: returns current file pointer position
    // -------------------------------------------------------------------------
    // Errors:          BT_FILE_ERROR_FILE_IS_NOT_OPEN,
    // BT_ERROR_SYSTEM *)
    function GetPosition(out fo_Pos: int64): TBTCommonError;
    /// /////////////////////////////////////////////////////////////////////////

    /// /////////////////////////////////////////////////////////////////////////
    // GetSize:         Get the size of the file
    // -------------------------------------------------------------------------
    // Parameters:      f_size: returns the current file size
    // -------------------------------------------------------------------------
    // Errors:          BT_FILE_ERROR_FILE_IS_NOT_OPEN,
    // BT_ERROR_SYSTEM *)
    function GetSize(out f_size: int64): TBTCommonError;
    /// /////////////////////////////////////////////////////////////////////////

    (* write to the file at the current file pointer position
      The file pointer will inc automaticly after writing *)
    function Write(const f_Data: Pointer; const f_Count: Cardinal;
      out fo_Written: Cardinal): TBTCommonError;

    (* read from file at the current file pointer position
      The file pointer will inc automaticly after writing *)
    function Read(out f_Data; f_Count: Cardinal; out fo_Read: Cardinal)
      : TBTCommonError;

    (* close an opened file
      function fails if not file is opened *)
    function CloseFile: TBTCommonError;

    (* returns an extended error information *)
    function GetErrorString(f_ErrorCode: TBTCommonError): string;
  end;

implementation

constructor TBTFileStream.Create;
begin
  m_FileIsOpen := False;
  m_LastErrorWasSystemError := False;
  inherited;
end;

destructor TBTFileStream.Destroy;
begin
  inherited;
end;

function TBTFileStream.OpenFileEx(const f_FilenameAnsi: PAnsiChar;
  const f_FilenameUnicode: PWideChar; const f_OpMode: TBTFileOpenMode;
  const f_OpParam: TBTFileModeParam;
  const f_ShMode: TBTFileShareMode = btfmShareNone): TBTCommonError;
var
  OpenParam, ShareMode: Cardinal;
begin
  if m_FileIsOpen then
  begin
    // a file is already opened
    Result := BT_FILE_ERROR_FILE_IS_OPEN;
    // mark last error as class error
    m_LastErrorWasSystemError := False;
    exit;
  end;

  // set parameters for open mode
  m_OpenMode := BT_FILE_ERROR_INVALID_FILE_OPEN_MODE;
  case f_OpMode of
    // share read mode
    btFmOpenRead:
      m_OpenMode := Windows.GENERIC_READ;
    // share write mode
    btFmOpenWrite:
      m_OpenMode := Windows.GENERIC_WRITE;
    // share read-write mode
    btFmOpenReadWrite:
      m_OpenMode := Windows.GENERIC_READ or Windows.GENERIC_WRITE;
  end;
  // check if a valid open parameter is set
  if m_OpenMode = BT_FILE_ERROR_INVALID_FILE_OPEN_MODE then
  begin
    Result := BT_FILE_ERROR_INVALID_FILE_OPEN_MODE;
    // mark last error as class error
    m_LastErrorWasSystemError := False;
    exit;
  end;

  // set ext. open parameters
  OpenParam := BT_FILE_ERROR_INVALID_FILE_OPEN_PARAM;
  case f_OpParam of
    btFmpCreateAlways:
      OpenParam := Windows.CREATE_ALWAYS;
    btFmpCreateNeweIfNotExist:
      OpenParam := Windows.CREATE_NEW;
    btFmpOpenAlways:
      OpenParam := Windows.OPEN_ALWAYS;
    btFmpOpenExisting:
      OpenParam := Windows.OPEN_EXISTING;
    btFmpTrunkateIfExist:
      OpenParam := Windows.TRUNCATE_EXISTING;
  end;
  // check if a valid ext. open parameter is set
  if OpenParam = BT_FILE_ERROR_INVALID_FILE_OPEN_PARAM then
  begin
    Result := BT_FILE_ERROR_INVALID_FILE_OPEN_PARAM;
    // mark last error as class error
    m_LastErrorWasSystemError := False;
    exit;
  end;

  // set parameters for share mode
  ShareMode := BT_FILE_ERROR_INVALID_FILE_SHARE_MODE;
  case f_ShMode of
    // share read mode
    btfmShareRead:
      ShareMode := Windows.FILE_SHARE_READ;
    // share write mode
    btfmShareWrite:
      ShareMode := Windows.FILE_SHARE_WRITE;
    // share read-write mode
    btfmShareReadWrite:
      ShareMode := Windows.FILE_SHARE_WRITE or Windows.FILE_SHARE_READ;
    // share none mode
    btfmShareNone:
      ShareMode := 0;
  end;
  // check if a valid fileshare parameter is set
  if ShareMode = BT_FILE_ERROR_INVALID_FILE_SHARE_MODE then
  begin
    Result := BT_FILE_ERROR_INVALID_FILE_SHARE_MODE;
    // mark last error as class error
    m_LastErrorWasSystemError := False;
    exit;
  end;

  (* Create the file:
    Share it for read operation of other processes, but denny them to write
    Create the file always with normal file attributes *)
  if f_FilenameAnsi <> nil then
    m_FileHandle := Windows.CreateFileA(f_FilenameAnsi, m_OpenMode, ShareMode,
      nil, OpenParam, FILE_ATTRIBUTE_NORMAL, 0)
  else
    m_FileHandle := Windows.CreateFileW(f_FilenameUnicode, m_OpenMode,
      ShareMode, nil, OpenParam, FILE_ATTRIBUTE_NORMAL, 0);

  // was creation successfull?
  if m_FileHandle = INVALID_HANDLE_VALUE then
  begin
    // an error occured
    m_LastWinAPIError := GetLastError;
    // try to translate the system error
    Result := SystemErrorToBTError(m_LastWinAPIError);
    // mark last error as system error
    m_LastErrorWasSystemError := True;
    // leave function
    exit;
  end;
  // everything seems fine
  m_FileIsOpen := True;
  Result := BT_OK;
end;

function TBTFileStream.OpenFile(const f_Filename: string;
  const f_OpMode: TBTFileOpenMode; const f_OpParam: TBTFileModeParam;
  const f_ShMode: TBTFileShareMode): TBTCommonError;
begin
{$IFDEF UNICODE}
  Result := OpenFileEx(nil, PWideChar(f_Filename), f_OpMode, f_OpParam,
    f_ShMode);
{$ELSE}
  Result := OpenFileEx(PAnsiChar(f_Filename), nil, f_OpMode, f_OpParam,
    f_ShMode);
{$ENDIF};

end;

function TBTFileStream.OpenFileA(const f_Filename: AnsiString;
  const f_OpMode: TBTFileOpenMode; const f_OpParam: TBTFileModeParam;
  const f_ShMode: TBTFileShareMode = btfmShareNone): TBTCommonError;
begin
  Result := OpenFileEx(PAnsiChar(f_Filename), nil, f_OpMode, f_OpParam,
    f_ShMode);
end;

function TBTFileStream.OpenFileW(const f_Filename: WideString;
  const f_OpMode: TBTFileOpenMode; const f_OpParam: TBTFileModeParam;
  const f_ShMode: TBTFileShareMode = btfmShareNone): TBTCommonError;
begin
  Result := OpenFileEx(nil, PWideChar(f_Filename), f_OpMode, f_OpParam,
    f_ShMode);
end;

const
  // used for Get and SetPosition error handling
  INVALID_SET_FILE_POINTER = DWORD(-1);

function TBTFileStream.SetPosition(const f_pos: int64): TBTCommonError;
var
  Low, High: DWORD;
  Size: int64;
begin
  if not m_FileIsOpen then
  begin
    // a file is already opened
    Result := BT_FILE_ERROR_FILE_IS_NOT_OPEN;
    // mark last error as class error
    m_LastErrorWasSystemError := False;
    exit;
  end;
  // get current size of the file
  Result := GetSize(Size);
  if Result <> BT_OK then
    exit;
  // check if desired position is possible
  if (f_pos > Size) or (f_pos < 0) then
  begin
    Result := BT_FILE_ERROR_INVALID_POSITION;
    // mark last error as class error
    m_LastErrorWasSystemError := False;
    exit;
  end;
  // set low DWORD for 64 bit filesize support
  Low := f_pos and $00000000FFFFFFFF;
  // set high DWORD for 64 bit filesize support
  High := (f_pos and $FFFFFFFF00000000) shr 32;
  if Windows.SetFilePointer(m_FileHandle, Low, @ high, FILE_BEGIN)
    = INVALID_SET_FILE_POINTER then
  begin
    // an error occured
    m_LastWinAPIError := GetLastError;
    // try to translate the system error
    Result := SystemErrorToBTError(m_LastWinAPIError);
    // mark last error as system error
    m_LastErrorWasSystemError := True;
    // leave function
    exit;
  end;
  Result := BT_OK;
end;

function TBTFileStream.GetPosition(out fo_Pos: int64): TBTCommonError;
var
  Low, High: DWORD;
begin
  if not m_FileIsOpen then
  begin
    // a file is already opened
    Result := BT_FILE_ERROR_FILE_IS_NOT_OPEN;
    // mark last error as class error
    m_LastErrorWasSystemError := False;
    exit;
  end;
  // do not change the position
  Low := $00000000;
  High := $00000000;
  // change posiiton with distance 0 from current ;)
  Low := Windows.SetFilePointer(m_FileHandle, Low, @ high, FILE_CURRENT);
  if Low = INVALID_SET_FILE_POINTER then
  begin
    // an error occured
    m_LastWinAPIError := GetLastError;
    // try to translate the system error
    Result := SystemErrorToBTError(m_LastWinAPIError);
    // mark last error as system error
    m_LastErrorWasSystemError := True;
    // leave function
    exit;
  end;
  // combine high and low dword to 64 bit value
  fo_Pos := (int64( High) shl 32) or int64( Low);
  // everthing seems fine
  Result := BT_OK;
end;

function TBTFileStream.GetSize(out f_size: int64): TBTCommonError;
var
  Low, High: DWORD;
begin
  if not m_FileIsOpen then
  begin
    // a file is already opened
    Result := BT_FILE_ERROR_FILE_IS_NOT_OPEN;
    // mark last error as class error
    m_LastErrorWasSystemError := False;
    exit;
  end;
  // get filesize (low and high dword)
  Low := GetFileSize(m_FileHandle, @ high);
  // check for error
  if Low = INVALID_FILE_SIZE then
  begin
    // an error occured
    m_LastWinAPIError := GetLastError;
    // try to translate the system error
    Result := SystemErrorToBTError(m_LastWinAPIError);
    // mark last error as system error
    m_LastErrorWasSystemError := True;
    // leave function
    exit;
  end;
  // combine high and low dword to 64 bit value
  f_size := (int64( High) shl 32) or int64( Low);
  // everthing seems fine
  Result := BT_OK;
end;

function TBTFileStream.Write(const f_Data: Pointer; const f_Count: Cardinal;
  out fo_Written: Cardinal): TBTCommonError;
begin
  if not m_FileIsOpen then
  begin
    // a file is already opened
    Result := BT_FILE_ERROR_FILE_IS_NOT_OPEN;
    // mark last error as class error
    m_LastErrorWasSystemError := False;
    exit;
  end;
  if m_OpenMode = GENERIC_READ then
  begin
    // write mode not supported
    Result := BT_FILE_ERROR_FILE_WRITE_NOT_SUPPORTED;
    // mark last error as class error
    m_LastErrorWasSystemError := False;
    exit;
  end;
  // Write the data to file (file pointer increase automaticly)
  if (WriteFile(m_FileHandle, f_Data^, f_Count, fo_Written, nil) = False) then
  begin
    // an error occured
    m_LastWinAPIError := GetLastError;
    // try to translate the system error
    Result := SystemErrorToBTError(m_LastWinAPIError);
    // mark last error as system error
    m_LastErrorWasSystemError := True;
    // leave function
    exit;
  end;
  // save curernt position
  Result := BT_OK;
end;

function TBTFileStream.Read(out f_Data; f_Count: Cardinal;
  out fo_Read: Cardinal): TBTCommonError;
begin
  if not m_FileIsOpen then
  begin
    // a file is already opened
    Result := BT_FILE_ERROR_FILE_IS_NOT_OPEN;
    // mark last error as class error
    m_LastErrorWasSystemError := False;
    exit;
  end;
  if m_OpenMode = GENERIC_WRITE then
  begin
    // write mode not supported
    Result := BT_FILE_ERROR_FILE_READ_NOT_SUPPORTED;
    // mark last error as class error
    m_LastErrorWasSystemError := False;
    exit;
  end;
  // Write the data from file (file pointer increase automaticly)
  if Windows.ReadFile(m_FileHandle, f_Data, f_Count, fo_Read, nil) = False then
  begin
    // an error occured
    m_LastWinAPIError := GetLastError;
    // try to translate the system error
    Result := SystemErrorToBTError(m_LastWinAPIError);
    // mark last error as system error
    m_LastErrorWasSystemError := True;
    // leave function
    exit;
  end;
  Result := BT_OK;
end;

function TBTFileStream.CloseFile: TBTCommonError;
begin
  if not m_FileIsOpen then
  begin
    // a file is already opened
    Result := BT_FILE_ERROR_FILE_IS_NOT_OPEN;
    // mark last error as class error
    m_LastErrorWasSystemError := False;
    exit;
  end;
  if Windows.CloseHandle(m_FileHandle) = False then
  begin
    // an error occured
    m_LastWinAPIError := GetLastError;
    // try to translate the system error
    Result := SystemErrorToBTError(m_LastWinAPIError);
    // mark last error as system error
    m_LastErrorWasSystemError := True;
    // leave function
    exit;
  end;
  m_FileIsOpen := False;
  Result := BT_OK;
end;

function TBTFileStream.GetErrorString(f_ErrorCode: TBTCommonError): string;
var
  l_sTemp: string;
begin
  Result := BT_FS_ERRSTR_UNKNOWN;
  if m_LastErrorWasSystemError then
    // checking for last system error
    Result := SystemErrorToString(m_LastWinAPIError)
  else
  begin
    // checking for class errors
    case f_ErrorCode of
      BT_FILE_ERROR_INVALID_POSITION:
        Result := BT_FS_ERRSTR_INVALID_POSITION;
      BT_FILE_ERROR_FILE_IS_OPEN:
        Result := BT_FS_ERRSTR_FILE_IS_OPEN;
      BT_FILE_ERROR_FILE_IS_NOT_OPEN:
        Result := BT_FS_ERRSTR_FILE_IS_NOT_OPEN;
      BT_FILE_ERROR_INVALID_FILE_OPEN_MODE:
        Result := BT_FS_ERRSTR_INVALID_FILE_OPEN_MODE;
      BT_FILE_ERROR_INVALID_FILE_OPEN_PARAM:
        Result := BT_FS_ERRSTR_INVALID_FILE_OPEN_PARAM;
      BT_FILE_ERROR_INVALID_FILE_SHARE_MODE:
        Result := BT_FS_ERRSTR_INVALID_FILE_SHARE_MODE;
      BT_FILE_ERROR_FILE_WRITE_NOT_SUPPORTED:
        Result := BT_FS_ERRSTR_FILE_WRITE_NOT_SUPPORTED;
      BT_FILE_ERROR_FILE_READ_NOT_SUPPORTED:
        Result := BT_FS_ERRSTR_FILE_READ_NOT_SUPPORTED;
    end;
  end;
  // Remove last line break if exist....
  if Length(Result) >= 2 then
  begin
    SetLength(l_sTemp, 2);
    CopyMemory(@l_sTemp[1], @Result[Length(Result) - 1], 2);
    if (l_sTemp = #10#13) or (l_sTemp = #13#10) then
    begin
      SetLength(l_sTemp, Length(Result) - 2);
      CopyMemory(@l_sTemp[1], @Result[1], Length(l_sTemp));
      Result := l_sTemp;
    end;
  end;
end;

function TBTFileStream.SystemErrorToBTError(f_SystemError: DWORD)
  : TBTCommonError;
begin
  Result := BT_ERROR_SYSTEM;
  case f_SystemError of
    // common Errors
    ERROR_REQUEST_ABORTED:
      Result := BT_COMMON_ERROR_REQUEST_ABORTED;
    // file Errors
    ERROR_FILE_NOT_FOUND:
      Result := BT_FILE_ERROR_FILE_NOT_FOUND;
    ERROR_PATH_NOT_FOUND:
      Result := BT_FILE_ERROR_PATH_NOT_FOUND;
    ERROR_TOO_MANY_OPEN_FILES:
      Result := BT_FILE_ERROR_TOO_MANY_OPEN_FILES;
    ERROR_WRITE_PROTECT:
      Result := BT_FILE_ERROR_WRITE_PROTECT;
    ERROR_SHARING_VIOLATION:
      Result := BT_FILE_ERROR_SHARING_VIOLATION;
    ERROR_SHARING_BUFFER_EXCEEDED:
      Result := BT_FILE_ERROR_SHARING_BUFFER_EXCEEDED;
    ERROR_LOCK_VIOLATION:
      Result := BT_FILE_ERROR_LOCK_VIOLATION;
    ERROR_HANDLE_DISK_FULL:
      Result := BT_FILE_ERROR_HANDLE_DISK_FULL;
    ERROR_FILE_EXISTS:
      Result := BT_FILE_ERROR_FILE_ALREADY_EXISTS;
    ERROR_OPEN_FAILED:
      Result := BT_FILE_ERROR_OPEN_FAILED;
    ERROR_BUFFER_OVERFLOW:
      Result := BT_FILE_ERROR_NAME_TOO_LONG;
    ERROR_DISK_FULL:
      Result := BT_FILE_ERROR_DISK_FULL;
    ERROR_NO_MORE_SEARCH_HANDLES:
      Result := BT_FILE_ERROR_NO_MORE_HANDLES;
    ERROR_INVALID_NAME:
      Result := BT_FILE_ERROR_INVALID_NAME;
    ERROR_ALREADY_EXISTS:
      Result := BT_FILE_ERROR_ALREADY_EXISTS_ON_CREATE;
    ERROR_FILENAME_EXCED_RANGE:
      Result := BT_FILE_ERROR_FILENAME_EXCED_RANGE;
    ERROR_META_EXPANSION_TOO_LONG:
      Result := BT_FILE_ERROR_META_EXPANSION;
    ERROR_FILE_CORRUPT:
      Result := BT_FILE_ERROR_FILE_CORRUPT;
    ERROR_DISK_CORRUPT:
      Result := BT_FILE_ERROR_DISK_CORRUPT;
    ERROR_ACCESS_DENIED:
      Result := BT_FILE_ERROR_ACCESS_DENIED;
    // network errors
    ERROR_NOT_SUPPORTED:
      Result := BT_NETWORK_ERROR_NETWORK_NOT_SUPPORTED;
    ERROR_BAD_NETPATH:
      Result := BT_NETWORK_ERROR_BAD_NETPATH;
    ERROR_NETWORK_BUSY:
      Result := BT_NETWORK_ERROR_NETWORK_BUSY;
  end;
end;

end.
